/*-
 * Copyright (c) 2010 Andriy Gapon <avg@FreeBSD.Org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/bio.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/dirent.h>
#include <sys/fcntl.h>
#include <sys/iconv.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/priv.h>
#include <sys/proc.h>
#include <sys/queue.h>
#include <sys/vnode.h>
#include <sys/endian.h>

#include <geom/geom.h>
#include <geom/geom_vfs.h>

#include <vm/uma.h>

#include "avg.h"

MALLOC_DEFINE(M_AVGFS, "avgfs", "AVG FS");

static vfs_init_t      avg_init;
static vfs_uninit_t    avg_uninit;
static vfs_mount_t     avg_mount;
static vfs_root_t      avg_root;
static vfs_statfs_t    avg_statfs;
static vfs_unmount_t   avg_unmount;
static vfs_fhtovp_t    avg_fhtovp;
/*static vfs_vget_t      avg_vget;*/

static struct vfsops avg_vfsops = {
	.vfs_fhtovp =		avg_fhtovp,
	.vfs_init =		avg_init,
	.vfs_mount =		avg_mount,
	.vfs_root =		avg_root,
	.vfs_statfs =		avg_statfs,
	.vfs_uninit =		avg_uninit,
	.vfs_unmount =		avg_unmount,
	.vfs_vget =		avg_vget,
};
VFS_SET(avg_vfsops, avg, VFCF_READONLY);

MODULE_VERSION(avg, 1);

static int avg_mountfs(struct vnode *, struct mount *);

static int
avg_init(struct vfsconf *foo)
{

	return (0);
}

static int
avg_uninit(struct vfsconf *foo)
{

	return (0);
}

static int
avg_mount(struct mount *mp)
{
	struct vnode *devvp;	/* vnode of the mount device */
	struct thread *td;
	struct vfsoptlist *opts;
	char *fspec;
	int error, len, use_devvp;
	struct nameidata nd, *ndp = &nd;

	td = curthread;
	opts = mp->mnt_optnew;

	/*
	 * Unconditionally mount as read-only.
	 */
	MNT_ILOCK(mp);
	mp->mnt_flag |= MNT_RDONLY;
	MNT_IUNLOCK(mp);

	/* No root filesystem support. */
	if (mp->mnt_flag & MNT_ROOTFS)
		return (ENOTSUP);

	use_devvp = (vfs_getopt(opts, "devvp", (void **)NULL, NULL) == 0);
	fspec = NULL;
	error = vfs_getopt(opts, "from", (void **)&fspec, &len);
	if (!error && fspec[len - 1] != '\0')
		return (EINVAL);

	if (mp->mnt_flag & MNT_UPDATE) {
		return (0);
	}

	/* Check that the mount device exists */
	if (fspec == NULL)
		return (EINVAL);
	NDINIT(ndp, LOOKUP, FOLLOW | LOCKLEAF, UIO_SYSSPACE, fspec, td);
	if ((error = namei(ndp)))
		return (error);
	NDFREE(ndp, NDF_ONLY_PNBUF);
	devvp = ndp->ni_vp;

	if (vn_isdisk(devvp, &error) == 0) {
		vput(devvp);
		return (error);
	}

	/* Check the access rights on the mount device */
	error = VOP_ACCESS(devvp, VREAD, td->td_ucred, td);
	if (error)
		error = priv_check(td, PRIV_VFS_MOUNT_PERM);
	if (error) {
		vput(devvp);
		return (error);
	}

	if ((error = avg_mountfs(devvp, mp))) {
		vrele(devvp);
		return (error);
	}
	VFSTOAVGFS(mp)->use_devvp = use_devvp;

	vfs_mountedfrom(mp, fspec);
	return (0);
};

static int
avg_mountfs(struct vnode *devvp, struct mount *mp)
{
	struct buf *bp = NULL;
	struct cdev *dev;
	struct avg_mnt *avgmp = NULL;
	int error = EINVAL;
	struct g_consumer *cp;
	struct bufobj *bo;

	dev = devvp->v_rdev;
	dev_ref(dev);
	DROP_GIANT();
	g_topology_lock();
	error = g_vfs_open(devvp, &cp, "avg", 0);
	g_topology_unlock();
	PICKUP_GIANT();
	VOP_UNLOCK(devvp, 0);
	if (error)
		goto bail;

	bo = &devvp->v_bufobj;

	if (devvp->v_rdev->si_iosize_max != 0)
		mp->mnt_iosize_max = devvp->v_rdev->si_iosize_max;
	if (mp->mnt_iosize_max > MAXPHYS)
		mp->mnt_iosize_max = MAXPHYS;

	avgmp = malloc(sizeof(struct avg_mnt), M_AVGFS,
	    M_WAITOK | M_ZERO);

	mp->mnt_data = avgmp;
	mp->mnt_stat.f_fsid.val[0] = dev2udev(devvp->v_rdev);
	mp->mnt_stat.f_fsid.val[1] = mp->mnt_vfc->vfc_typenum;
	MNT_ILOCK(mp);
	mp->mnt_flag |= MNT_LOCAL;
	mp->mnt_kern_flag |= MNTK_MPSAFE | MNTK_LOOKUP_SHARED |
	    MNTK_EXTENDED_SHARED;
	MNT_IUNLOCK(mp);
	avgmp->im_mountp = mp;
	avgmp->im_dev = dev;
	avgmp->im_devvp = devvp;
	avgmp->im_cp = cp;
	avgmp->im_bo = bo;
	avgmp->size = cp->provider->mediasize;
	avgmp->bsize = cp->provider->sectorsize;
	avgmp->bmask = avgmp->bsize - 1;
	avgmp->bshift = ffs(avgmp->bsize) - 1;

	return (0);

bail:
	if (avgmp != NULL)
		free(avgmp, M_AVGFS);
	if (bp != NULL)
		brelse(bp);
	if (cp != NULL) {
		DROP_GIANT();
		g_topology_lock();
		g_vfs_close(cp);
		g_topology_unlock();
		PICKUP_GIANT();
	}
	dev_rel(dev);
	return error;
};

static int
avg_unmount(struct mount *mp, int mntflags)
{
	struct avg_mnt *avgmp;
	int error, flags = 0;

	avgmp = VFSTOAVGFS(mp);

	if (mntflags & MNT_FORCE)
		flags |= FORCECLOSE;

	if ((error = vflush(mp, 0, flags, curthread)))
		return (error);

	DROP_GIANT();
	g_topology_lock();
	g_vfs_close(avgmp->im_cp);
	g_topology_unlock();
	PICKUP_GIANT();
	vrele(avgmp->im_devvp);
	dev_rel(avgmp->im_dev);

	free(avgmp, M_AVGFS);
	mp->mnt_data = NULL;
	MNT_ILOCK(mp);
	mp->mnt_flag &= ~MNT_LOCAL;
	MNT_IUNLOCK(mp);

	return (0);
}

static int
avg_root(struct mount *mp, int flags, struct vnode **vpp)
{
	ino_t id;

	id = 1;
	return (avg_vget(mp, id, flags, vpp));
}

static int
avg_statfs(struct mount *mp, struct statfs *sbp)
{
	struct avg_mnt *avgmp;

	avgmp = VFSTOAVGFS(mp);

	sbp->f_bsize = avgmp->bsize;
	sbp->f_iosize = avgmp->bsize;
	sbp->f_blocks = avgmp->size >> avgmp->bshift;
	sbp->f_bfree = 0;
	sbp->f_bavail = 0;
	sbp->f_files = 0;
	sbp->f_ffree = 0;
	return 0;
}

int
avg_vget(struct mount *mp, ino_t ino, int flags, struct vnode **vpp)
{
	struct avg_mnt *avgmp;
	struct thread *td;
	struct vnode *vp;
	struct avg_node *unode;
	int error;

	error = vfs_hash_get(mp, ino, flags, curthread, vpp, NULL, NULL);
	if (error || *vpp != NULL)
		return (error);

	/*
	 * We must promote to an exclusive lock for vnode creation.  This
	 * can happen if lookup is passed LOCKSHARED.
 	 */
	if ((flags & LK_TYPE_MASK) == LK_SHARED) {
		flags &= ~LK_TYPE_MASK;
		flags |= LK_EXCLUSIVE;
	}

	/*
	 * We do not lock vnode creation as it is believed to be too
	 * expensive for such rare case as simultaneous creation of vnode
	 * for same ino by different processes. We just allow them to race
	 * and check later to decide who wins. Let the race begin!
	 */
	td = curthread;
	if ((error = avg_allocv(mp, &vp, td))) {
		printf("Error from udf_allocv\n");
		return (error);
	}

	avgmp = VFSTOAVGFS(mp);
	unode = malloc(sizeof(struct avg_node), M_AVGFS, M_WAITOK | M_ZERO);
	unode->i_vnode = vp;
	unode->hash_id = ino;
	unode->avgmp = avgmp;
	vp->v_data = unode;

	lockmgr(vp->v_vnlock, LK_EXCLUSIVE, NULL);
	error = insmntque(vp, mp);
	if (error != 0)
		return (error);

	error = vfs_hash_insert(vp, ino, flags, td, vpp, NULL, NULL);
	if (error || *vpp != NULL)
		return (error);

	switch (ino) {
	default:
		vp->v_type = VBAD;
		break;
	case AVG_ROOTDIR_INO:
		vp->v_type = VDIR;
		break;
	case AVG_THEFILE_INO:
		vp->v_type = VREG;
		break;
	}

	if (vp->v_type != VFIFO)
		VN_LOCK_ASHARE(vp);

	if (ino == AVG_ROOTDIR_INO)
		vp->v_vflag |= VV_ROOT;

	*vpp = vp;

	return (0);
}

static int
avg_fhtovp(struct mount *mp, struct fid *fhp, struct vnode **vpp)
{
	struct ifid *ifhp;
	struct vnode *nvp;
	off_t fsize;
	int error;

	ifhp = (struct ifid *)fhp;

	if ((error = VFS_VGET(mp, ifhp->ifid_ino, LK_EXCLUSIVE, &nvp)) != 0) {
		*vpp = NULLVP;
		return (error);
	}

	fsize = VTON(nvp)->avgmp->size;

	*vpp = nvp;
	vnode_create_vobject(*vpp, fsize, curthread);
	return (0);
}

